# __init__.py
# Copyright 2020 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a NoSQL database of chess games using the dbm.gnu package.

This interface can be used if GNU's dbm software is installed and available to
the dbm.gnu package.
"""



# selectionds.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Interface to chess database for selection rules index.

The SelectionDS class in this module supports the apsw, db, and sqlite3,
interfaces to a database.

See the ..dpt.selectionds module for the SelectionDS class for DPT.

"""

from solentware_grid.core.datasourcecursor import DataSourceCursor

from .selection import Selection


class SelectionDS(Selection, DataSourceCursor):

    """Combine a standard DataSourceCursor with Selection."""



# chessgame.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Demonstrate chess game class and methods to display PGN text, board, and
analysis as it appears without an active chess engine."""

if __name__ == "__main__":

    import tkinter

    from pgn_read.core.parser import PGN

    from ..gui import fonts
    from ..gui.game import Game
    from .chessscore import DuckUI

    class DuckStatusbar:
        """Helper to allow Game instances to work without a real UI."""

        def set_status_text(self, *a, **k):
            pass

    class DuckqueueQueue:
        """Helper to allow Game instances to work without a real UI."""

        def put(self, *a):
            pass

    class DuckcoreuciUCI:
        """Helper to allow Game instances to work without a real UI."""

        ui_analysis_queue = DuckqueueQueue()
        position_analysis = {}

    class DuckguiuciUCI:
        """Helper to allow Game instances to work without a real UI."""

        uci = DuckcoreuciUCI()

    class DuckUI(DuckUI):
        """Helper to allow Game instances to work without a real UI."""

        show_analysis = True
        visible_scrollbars = True

        # None is legal but the default analysis does not get shown in the
        # analysis widget.
        database = False

        def make_position_analysis_data_source(self):
            return None

        statusbar = DuckStatusbar()
        uci = DuckguiuciUCI()

    # Maybe the real generate_popup_navigation_maps can be declared somewhere
    # else, even if duplicated, to avoid this.
    # An empty Navigation entry is put in the popup menu.
    class Game(Game):
        binding_labels = ()

        def generate_popup_navigation_maps(self):
            return {}, {}

    root = tkinter.Tk()
    root.wm_minsize(width=900, height=600)
    f = fonts.make_chess_fonts(root)
    root.wm_title("Demonstrate Game")
    root.pack_propagate(False)

    # Use DuckUI.
    g = Game(master=root, ui=DuckUI(root))

    del f
    g.collected_game = next(
        PGN().read_games(
            "".join(
                (
                    '[Event"National Club: Gosport - Wood Green"]',
                    '[Site"Gosport"]',
                    '[Date"1989-05-07"]',
                    '[Round"QFinal"]',
                    '[White"Sowray P J"]',
                    '[Black"Marsh R"]',
                    '[Result"1-0"]',
                    "e4(d4d5c4e6Nc3)c6d4d5exd5cxd5c4Nf6c5e6Nc3b6b4a5Bf4",
                    "axb4Nb5Na6Qa4;comment to eol\nBd7",
                    "Bc7Nxc5Qd1Qc8dxc5bxc5Nf3Qb7Nd6Bxd6Bxd6Qb6Be5Ke7Be2Ne4O-Of6Bb2Nc3",
                    "Bxc3bxc3Qd3(Qb3)Ra3Rfb1Qa7Qc2g6Rb3d4Bc4Rxb3Bxb3Qa6a4Rb8a5e5Bd5",
                    "\n%The escape sequence\nRb2",
                    "Qe4Bf5Qh4Qd3(c2(g5)Nd2Qxa5Rxa5Rb1{Comment\ncontaining newline}Nf1",
                    "(Nxb1c1=Q)Rxf1Kxf1Bd3)g4Rb1",
                    "Rxb1Qxb1Kg2Kd6Qxf6Kxd5<reserved\n\nfor future use>Qxe5",
                    "Kc6gxf5Qxf5Qe8Kc7Qe7Kc8Ne5c2Qxc5Kd8Qxd4",
                    "Ke8Qe3Kf8Kg3Qc8Nd3Kg8f4Qc6Nc1Qa4Qb3",
                    "1-0",
                    "\n",
                )
            )
        )
    )

    g.set_and_tag_item_text()

    # The get_top_widget method and takefocus_widget attribute are defined in
    # Game class to keep them out of AnalysisScore instances.
    g.get_top_widget().pack(fill=tkinter.BOTH, expand=tkinter.TRUE)
    g.score.focus_set()
    del g

    root.mainloop()



# chessrecord.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Record definitions for chess game database.

The ...Game... classes differ in the PGN parser used as a superclass of the
...valueGame... class.  These generate different combinations of the available
data structures from the game score for the various display and update uses.
The ...Update classes allow editing of a, possibly incomplete, game score.

"""
from ast import literal_eval
import re

from solentware_base.core.record import KeyData, Value, ValueText, Record
from solentware_base.core.segmentsize import SegmentSize

from pgn_read.core.parser import PGN
from pgn_read.core.constants import (
    SEVEN_TAG_ROSTER,
    TAG_DATE,
    TAG_WHITE,
    TAG_BLACK,
)

from .pgn import (
    GameDisplayMoves,
    GameRepertoireDisplayMoves,
    GameRepertoireTags,
    GameRepertoireUpdate,
    GameTags,
    GameUpdate,
)
from .constants import (
    START_RAV,
    END_RAV,
    NON_MOVE,
    TAG_OPENING,
    START_COMMENT,
    ERROR_START_COMMENT,
    ESCAPE_END_COMMENT,
    HIDE_END_COMMENT,
    END_COMMENT,
    SPECIAL_TAG_DATE,
)
from .cqlstatement import CQLStatement
from .filespec import (
    POSITIONS_FIELD_DEF,
    SOURCE_FIELD_DEF,
    PIECESQUAREMOVE_FIELD_DEF,
    PIECEMOVE_FIELD_DEF,
    SQUAREMOVE_FIELD_DEF,
    GAMES_FILE_DEF,
    REPERTOIRE_FILE_DEF,
    OPENING_ERROR_FIELD_DEF,
    PGN_DATE_FIELD_DEF,
    VARIATION_FIELD_DEF,
    ENGINE_FIELD_DEF,
    PARTIALPOSITION_NAME_FIELD_DEF,
    RULE_FIELD_DEF,
    COMMAND_FIELD_DEF,
)
from .analysis import Analysis
from .querystatement import QueryStatement, re_normalize_player_name
from .engine import Engine

PLAYER_NAME_TAGS = frozenset((TAG_WHITE, TAG_BLACK))


class ChessRecordError(Exception):
    pass


class ChessDBkeyGame(KeyData):

    """Primary key of chess game."""

    def __eq__(self, other):
        """Return (self == other).  Attributes are compared explicitly."""
        try:
            return self.recno == other.recno
        except:
            return False

    def __ne__(self, other):
        """Return (self != other).  Attributes are compared explicitly."""
        try:
            return self.recno != other.recno
        except:
            return True


class ChessDBvaluePGN(Value):

    """Methods common to all chess PGN data classes."""

    def __init__(self):
        super().__init__()
        self.collected_game = None

    @staticmethod
    def encode_move_number(key):
        """Return base 256 string for integer with left-end most significant."""
        return key.to_bytes(2, byteorder="big")

    def load(self, value):
        """Get game from value."""
        self.collected_game = next(self.read_games(literal_eval(value)))

    def pack_value(self):
        """Return PGN text for game."""
        return repr("".join(self.collected_game._text))


class ChessDBvalueGame(PGN, ChessDBvaluePGN):

    """Chess game data.

    Data is indexed by PGN Seven Tag Roster tags.

    """

    def __init__(self, game_class=GameDisplayMoves):
        """Extend with game source and move number encoder placeholders."""
        super().__init__(game_class=game_class)

    def pack(self):
        """Return PGN text and indexes for game."""
        v = super(ChessDBvalueGame, self).pack()
        index = v[1]
        tags = self.collected_game._tags
        for field in tags:
            if field in PLAYER_NAME_TAGS:

                # PGN specification states colon is used to separate player
                # names in consultation games.
                index[field] = [
                    " ".join(re_normalize_player_name.findall(tf))
                    for tf in tags[field].split(":")
                ]

            elif field in SEVEN_TAG_ROSTER:
                index[field] = [tags[field]]
        if TAG_DATE in tags:
            index[PGN_DATE_FIELD_DEF] = [
                tags[TAG_DATE].replace(*SPECIAL_TAG_DATE)
            ]
        return v


class ChessDBrecordGame(Record):

    """Chess game record customised for displaying the game score and tags."""

    def __init__(self):
        """Extend with move number encode and decode methods"""
        super(ChessDBrecordGame, self).__init__(
            ChessDBkeyGame, ChessDBvalueGame
        )

    def clone(self):
        """Return copy of ChessDBrecordGame instance.

        The bound method attributes are dealt with explicitly and the rest
        are handled by super(...).clone().  (Hope that DPT CR LF restrictions
        will be removed at which point the bound method attributes will not
        be needed.  Then ChessDBrecordGame.clone() can be deleted.)

        """
        # are conditions for deleting this method in place?
        clone = super(ChessDBrecordGame, self).clone()
        return clone

    @staticmethod
    def decode_move_number(skey):
        """Return integer from base 256 string with left-end most significant."""
        return int.from_bytes(skey, byteorder="big")

    def delete_record(self, database, dbset):
        """Delete record not allowed using ChessDBrecordGame class.

        Process the game using a ChessDBrecordGameUpdate instance

        """
        raise ChessRecordError

    def edit_record(self, database, dbset, dbname, newrecord):
        """Edit record not allowed using ChessDBrecordGame class.

        Process the game using a ChessDBrecordGameUpdate instance

        """
        raise ChessRecordError

    def get_keys(self, datasource=None, partial=None):
        """Return list of (key, value) tuples.

        The keys for the secondary databases in a ChessDatabase instance are
        embedded in, or derived from, the PGN string for the game.  All
        except the positions are held in self.value.collected_game._tags.
        Multiple position keys can arise becuse all repetitions of a
        position are of interest.  The partial argument is used to select
        the relevant keys.  The position is in partial and the keys will
        differ in the move number.

        """
        dbname = datasource.dbname
        if dbname != POSITIONS_FIELD_DEF:
            if dbname == GAMES_FILE_DEF:
                return [(self.key.recno, self.srvalue)]
            elif dbname in self.value.collected_game._tags:
                return [
                    (self.value.collected_game._tags[dbname], self.key.pack())
                ]
            else:
                return []
        if partial == None:
            return []

        moves = self.value.moves
        gamekey = datasource.dbhome.encode_record_number(self.key.pack())
        rav = 0
        ref = 0
        keys = []
        convert_format = datasource.dbhome.db_compatibility_hack

        p = tuple(partial)
        for mt in moves:
            if mt == START_RAV:
                rav += 1
            elif mt == END_RAV:
                rav -= 1
            elif mt == NON_MOVE:
                pass
            else:
                if mt[-1] == p:
                    record = (partial, None)
                    keys.append(convert_format(record, gamekey))
            ref += 1
        return keys

    def put_record(self, database, dbset):
        """Put record not allowed using ChessDBrecordGame class.

        Process the game using a ChessDBrecordGameUpdate instance

        """
        raise ChessRecordError


class ChessDBrecordGameText(Record):

    """Chess game record customised for processing the game score as text.

    Used to export games or repertoires from a database in the 'Import Format',
    see PGN specification 3.1, used to store the games.

    """

    def __init__(self):
        """Extend with move number encode and decode methods"""
        super(ChessDBrecordGameText, self).__init__(ChessDBkeyGame, ValueText)

    def clone(self):
        """Return copy of ChessDBrecordGameText instance.

        The bound method attributes are dealt with explicitly and the rest
        are handled by super(...).clone().  (Hope that DPT CR LF restrictions
        will be removed at which point the bound method attributes will not
        be needed.  Then ChessDBrecordGameText.clone() can be deleted.)

        """
        # are conditions for deleting this method in place?
        clone = super(ChessDBrecordGameText, self).clone()
        return clone

    @staticmethod
    def decode_move_number(skey):
        """Return integer from base 256 string with left-end most significant."""
        return int.from_bytes(skey, byteorder="big")

    def delete_record(self, database, dbset):
        """Delete record not allowed using ChessDBrecordGameText class.

        Process the game using a ChessDBrecordGameUpdate instance

        """
        raise ChessRecordError

    def edit_record(self, database, dbset, dbname, newrecord):
        """Edit record not allowed using ChessDBrecordGameText class.

        Process the game using a ChessDBrecordGameUpdate instance

        """
        raise ChessRecordError

    def put_record(self, database, dbset):
        """Put record not allowed using ChessDBrecordGameText class.

        Process the game using a ChessDBrecordGameUpdate instance

        """
        raise ChessRecordError


class ChessDBvalueGameTags(ChessDBvalueGame):

    """Chess game data excluding PGN movetext but including PGN Tags."""

    def __init__(self):
        """Extend with game source and move number encoder placeholders."""
        super().__init__(game_class=GameTags)

    def get_field_value(self, fieldname, occurrence=0):
        """Return value of a field occurrence, the first by default.

        Added to support Find and Where classes.

        """
        return self.collected_game._tags.get(fieldname, None)

    # def get_field_values(self, fieldname):
    #    """Return tuple of field values for fieldname.

    #    Added to support Find and Where classes.

    #    """
    #    return self.get_field_value(fieldname),

    def load(self, value):
        """Get game from value.

        The exception is a hack introduced to cope with a couple of games
        found in TWIC downloads which give the result as '1-0 ff' in the
        Result tag, and append ' ff' to the movetext after the '1-0' Game
        Termination Marker.  The 'ff' gets stored on ChessTab databases as
        the game score for an invalid game score.

        It is assumed other cases will need this trick, which seems to be
        needed only when displaying a list of games and not when displaying
        the full game score.

        """
        try:
            super().load(value)
        except StopIteration:
            self.collected_game = next(
                self.read_games("{" + literal_eval(value) + "}*")
            )


class ChessDBrecordGameTags(Record):

    """Chess game record customised for displaying tag information for a game."""

    def __init__(self):
        """Extend with move number encode and decode methods"""
        super(ChessDBrecordGameTags, self).__init__(
            ChessDBkeyGame, ChessDBvalueGameTags
        )


class ChessDBrecordGamePosition(Record):

    """Chess game record customised for displaying the game score only.

    Much of the game structure to be represented in the row display is held
    in the Tkinter.Text object created for display.  Thus the processing of
    the record data is delegated to a PositionScore instance created when
    filling the grid.

    """

    def __init__(self):
        """Extend with move number encode and decode methods"""
        super(ChessDBrecordGamePosition, self).__init__(
            ChessDBkeyGame, ValueText
        )

    def clone(self):
        """Return copy of ChessDBrecordGamePosition instance.

        The bound method attributes are dealt with explicitly and the rest
        are handled by super(...).clone().  (Hope that DPT CR LF restrictions
        will be removed at which point the bound method attributes will not
        be needed.  Then ChessDBrecordGamePosition.clone() can be deleted.)

        """
        # are conditions for deleting this method in place?
        clone = super(ChessDBrecordGamePosition, self).clone()
        return clone

    @staticmethod
    def decode_move_number(skey):
        """Return integer from base 256 string with left-end most significant."""
        return int.from_bytes(skey, byteorder="big")

    def delete_record(self, database, dbset):
        """Delete record not allowed using ChessDBrecordGamePosition class.

        Process the game using a ChessDBrecordGameUpdate instance

        """
        raise ChessRecordError

    def edit_record(self, database, dbset, dbname, newrecord):
        """Edit record not allowed using ChessDBrecordGamePosition class.

        Process the game using a ChessDBrecordGameUpdate instance

        """
        raise ChessRecordError

    def put_record(self, database, dbset):
        """Put record not allowed using ChessDBrecordGamePosition class.

        Process the game using a ChessDBrecordGameUpdate instance

        """
        raise ChessRecordError


# An alternative is to put this in .core.pgn as 'class _Game(Game) and base
# the other classes on _Game rather than Game.  Done here because the only
# place errors which need hiding should occur is when importing games to, or
# updating games on, the database.
class _GameUpdate(GameUpdate):
    """Override the PGN error notification and recovery methods.

    Errors detected in PGN movetext are hidden by wrapping all tokens to end
    of variation, which may be rest of game if error is in main line, in a
    comment which starts and ends with a presumed unlikely character sequence.
    The '}' in any '{}' comments which happen to get wrapped a changed to a
    distinct presumed unlikely character sequence so the wrapped '}' tokens do
    not terminate the wrapping comment.

    """

    def pgn_error_notification(self):
        """Insert error '{' before movetext token which causes PGN error."""
        if self._movetext_offset is not None:
            self._text.append(START_COMMENT + ERROR_START_COMMENT)

    def pgn_error_recovery(self):
        """Insert error '}' before token which ends the scope of a PGN error.

        This token will be a ')' or one of the game termination markers.

        """
        if self._movetext_offset is not None:
            self._text.append(ESCAPE_END_COMMENT + END_COMMENT)

    def pgn_mark_comment_in_error(self, comment):
        """Return comment with '}' replaced by a presumed unlikely sequence.

        One possibility is to wrap the error in a '{...}' comment.  The '}'
        token in any wrapped commment would end the comment wrapping the error
        prematurely, so replace with HIDE_END_COMMENT.

        """
        return comment.replace(END_COMMENT, HIDE_END_COMMENT)


class ChessDBvaluePGNUpdate(PGN, ChessDBvaluePGN):

    """Chess game data with position, piece location, and PGN Tag, indexes."""

    # Replaces ChessDBvaluePGNUpdate and ChessDBvalueGameImport which had been
    # identical for a considerable time.
    # Decided that PGNUpdate should remain in pgn.core.parser because that code
    # generates data while this code updates a database.
    # Now moved to .core.pgn.GameUpdate.
    # ChessDBvalueGameImport had this comment:
    # Implication of original is encode_move_number not supported and load in
    # ChessDBvaluePGN superclass is used.

    def __init__(self):
        """Extend with game source and move number encoder placeholders."""
        super().__init__(game_class=_GameUpdate)
        self.gamesource = None

    # Perhaps ChessDBvaluePGNUpdate should follow example of ChessDBvalueGame
    # in handling Seven Tag Roster, and just use 'try ... except ...' for
    # bulk imports in ChessDBrecordGameImport, where Seven Tag Roster should
    # be present.  This would need a subclass of ChessDBvaluePGNUpdate to hold
    # the version of pack() below.
    def pack(self):
        """Return PGN text and indexes for game."""
        v = super().pack()
        index = v[1]
        cg = self.collected_game
        if self.do_full_indexing():
            tags = cg._tags
            for field in SEVEN_TAG_ROSTER:
                if field in PLAYER_NAME_TAGS:

                    # PGN specification states colon is used to separate player
                    # names in consultation games.
                    try:
                        index[field] = [
                            " ".join(re_normalize_player_name.findall(tf))
                            for tf in tags[field].split(":")
                        ]
                    except KeyError:
                        if field in tags:
                            raise

                else:
                    try:
                        index[field] = [tags[field]]
                    except KeyError:
                        if field in tags:
                            raise
            index[POSITIONS_FIELD_DEF] = cg.positionkeys
            index[PIECESQUAREMOVE_FIELD_DEF] = cg.piecesquaremovekeys
            index[PIECEMOVE_FIELD_DEF] = cg.piecemovekeys
            index[SQUAREMOVE_FIELD_DEF] = cg.squaremovekeys
            try:
                index[PGN_DATE_FIELD_DEF] = [
                    tags[TAG_DATE].replace(*SPECIAL_TAG_DATE)
                ]
            except KeyError:
                if TAG_DATE in tags:
                    raise
        else:
            index[SOURCE_FIELD_DEF] = [self.gamesource]
        return v

    def set_game_source(self, source):
        """Set the index value to use if full indexing is not to be done."""
        self.gamesource = source

    def do_full_indexing(self):
        """Return True if full indexing is to be done.

        Detected PGN errors are wrapped in a comment starting 'Error: ' so
        method is_pgn_valid() is not used to decide what indexing to do.
        """
        return self.gamesource is None

    # Before ChessTab 4.3 could test the string attribute of any re.match
    # object for PGN text.  The match objects are not available in version
    # 4.3 and later.  At 4.3 the test was done only for the value attribute
    # of a record, so it is possible to test against the srvalue attribute
    # of the record instance.
    def is_error_comment_present(self):
        """Return True if an {Error: ...} comment is in the PGN text."""
        return START_COMMENT + ERROR_START_COMMENT in "".join(
            self.collected_game._text
        )


class ChessDBrecordGameUpdate(Record):

    """Chess game record customized for editing database records.

    Used to edit or insert a single record by typing in a widget.

    """

    def __init__(self):
        """Extend with move number encode and decode methods"""
        super(ChessDBrecordGameUpdate, self).__init__(
            ChessDBkeyGame, ChessDBvaluePGNUpdate
        )

    def clone(self):
        """Return copy of ChessDBrecordGameUpdate instance.

        The bound method attributes are dealt with explicitly and the rest
        are handled by super(...).clone().  (Hope that DPT CR LF restrictions
        will be removed at which point the bound method attributes will not
        be needed.  Then ChessDBrecordGameUpdate.clone() can be deleted.)

        """
        # are conditions for deleting this method in place?
        clone = super(ChessDBrecordGameUpdate, self).clone()
        return clone

    @staticmethod
    def decode_move_number(skey):
        """Return integer from base 256 string with left-end most significant."""
        return int.from_bytes(skey, byteorder="big")

    def get_keys(self, datasource=None, partial=None):
        """Return list of (key, value) tuples.

        The keys for the secondary databases in a ChessDatabase instance are
        embedded in, or derived from, the PGN string for the game.  All
        except the positions are held in self.value.collected_game._tags.
        Multiple position keys can arise becuse all repetitions of a
        position are of interest.  The partial argument is used to select
        the relevant keys.  The position is in partial and the keys will
        differ in the move number.

        """
        dbname = datasource.dbname
        if dbname != POSITIONS_FIELD_DEF:
            if dbname == GAMES_FILE_DEF:
                return [(self.key.recno, self.srvalue)]
            elif dbname in self.value.collected_game._tags:
                return [
                    (self.value.collected_game._tags[dbname], self.key.pack())
                ]
            else:
                return []
        if partial == None:
            return []

        moves = self.value.moves
        gamekey = datasource.dbhome.encode_record_number(self.key.pack())
        rav = 0
        ref = 0
        keys = []
        convert_format = datasource.dbhome.db_compatibility_hack

        p = tuple(partial)
        for mt in moves:
            if mt == START_RAV:
                rav += 1
            elif mt == END_RAV:
                rav -= 1
            elif mt == NON_MOVE:
                pass
            else:
                if mt[-1] == p:
                    record = (partial, None)
                    keys.append(convert_format(record, gamekey))
            ref += 1
        return keys

    def load_instance(self, database, dbset, dbname, record):
        """Extend to set source for game if necessary."""
        super(ChessDBrecordGameUpdate, self).load_instance(
            database, dbset, dbname, record
        )

        # Never called because attribute is not bound anywhere and no
        # exceptions are seen ever.
        # Until ..tools.chesstab-4-1-1_castling-option-correction written, and
        # following code commented.  I assume the idea made sense once!
        # Might as well let superclass method be used directly.
        # if self.value.callbacktried:
        #    pass
        # elif self.value.callbacktried == None:
        #    pass
        # elif not self.value.callbacktried:
        #    self.value.set_game_source(record[0])


class ChessDBrecordGameImport(Record):

    """Chess game record customised for importing games from PGN files.

    Used to import multiple records from a PGN file.

    """

    def __init__(self):
        """Customise Record with chess database import key and value classes"""
        super(ChessDBrecordGameImport, self).__init__(
            ChessDBkeyGame, ChessDBvaluePGNUpdate
        )

    def import_pgn(self, database, source, sourcename, reporter=None):
        """Update database with games read from source."""
        self.set_database(database)
        # self.value.set_game_source(sourcename)
        if reporter is not None:
            reporter("Extracting games from " + sourcename)
        ddup = database.deferred_update_points
        db_segment_size = SegmentSize.db_segment_size
        value = self.value
        count = 0
        for collected_game in value.read_games(source):
            value.set_game_source(
                sourcename if not collected_game.is_pgn_valid() else None
            )
            self.key.recno = None
            value.collected_game = collected_game
            self.put_record(self.database, GAMES_FILE_DEF)
            if reporter is not None:
                if not count:
                    base = self.key.recno - self.key.recno % db_segment_size
                count += 1
                if self.key.recno % db_segment_size in ddup:
                    reporter(
                        " ".join(
                            (
                                "Game",
                                str(count),
                                "stored as record",
                                str(self.key.recno),
                                "offset",
                                str(self.key.recno - base),
                            )
                        )
                    )
                    database.deferred_update_housekeeping()
        if reporter is not None:
            reporter("Extraction from " + sourcename + " done")
            reporter("", timestamp=False)


class ChessDBkeyPartial(KeyData):

    """Primary key of partial position record."""


class ChessDBvaluePartial(CQLStatement, Value):

    """Partial position data."""

    def __init__(self):
        super(ChessDBvaluePartial, self).__init__()

    def __eq__(self, other):
        """Return (self == other).  Attributes are compared explicitly."""
        try:
            if (
                self.get_name_statement_text()
                != other.get_name_statement_text()
            ):
                return False
            else:
                return True
        except:
            return False

    def __ne__(self, other):
        """Return (self != other).  Attributes are compared explicitly."""
        try:
            if (
                self.get_name_statement_text()
                == other.get_name_statement_text()
            ):
                return False
            else:
                return True
        except:
            return True

    def load(self, value):
        """Set partial position from value"""
        self.process_statement(literal_eval(value))

    def pack_value(self):
        """Return partial position value"""
        return repr(self.get_name_statement_text())

    def pack(self):
        """Extend, return partial position record and index data."""
        v = super().pack()
        index = v[1]
        index[PARTIALPOSITION_NAME_FIELD_DEF] = [self.get_name_text()]
        return v


class ChessDBrecordPartial(Record):

    """Partial position record."""

    def __init__(self):
        """Extend as a partial position record."""
        super(ChessDBrecordPartial, self).__init__(
            ChessDBkeyPartial, ChessDBvaluePartial
        )

    def get_keys(self, datasource=None, partial=None):
        """Return list of (key, value) tuples.

        The partial position name is held in an attribute which is not named
        for the field where it exists in the database.

        """
        if datasource.dbname == PARTIALPOSITION_NAME_FIELD_DEF:
            return [(self.value.get_name_text(), self.key.pack())]
        else:
            return super().get_keys(datasource=datasource, partial=partial)

    def load_value(self, value):
        """Load self.value from value which is repr(<data>).

        Set database in self.value for query processing then delegate value
        processing to superclass.

        """
        self.value.set_database(self.database)
        self.value.dbset = GAMES_FILE_DEF
        super().load_value(value)


# Not quite sure what customization needed yet
class ChessDBvalueRepertoire(PGN, Value):

    """Repertoire data using custom non-standard tags in PGN format."""

    def __init__(self, game_class=GameRepertoireDisplayMoves):
        """Extend with game source and move number encoder placeholders."""
        super().__init__(game_class=game_class)
        self.collected_game = None

    def load(self, value):
        """Get game from value."""
        self.collected_game = next(self.read_games(literal_eval(value)))


# Not quite sure what customization needed yet
class ChessDBvalueRepertoireTags(ChessDBvalueRepertoire):

    """Repertoire data using custom non-standard tags in PGN format."""

    def __init__(self):
        """Extend with game source and move number encoder placeholders."""
        super().__init__(game_class=GameRepertoireTags)


# Not quite sure what customization needed yet
class ChessDBvalueRepertoireUpdate(PGN, ChessDBvaluePGN):

    """Repertoire data using custom non-standard tags in PGN format."""

    def __init__(self):
        """Extend with game source and move number encoder placeholders."""
        super().__init__(game_class=GameRepertoireUpdate)
        self.gamesource = None

    def pack(self):
        """Return PGN text and indexes for game."""
        v = super(ChessDBvalueRepertoireUpdate, self).pack()
        index = v[1]
        tags = self.collected_game._tags
        if self.collected_game.is_pgn_valid():
            index[TAG_OPENING] = [tags[TAG_OPENING]]
        elif tags[TAG_OPENING]:
            index[TAG_OPENING] = [tags[TAG_OPENING]]
        else:
            index[TAG_OPENING] = [self.gamesource]
        return v

    def set_game_source(self, source):
        """Set game source (the PGN file name or '')"""
        self.gamesource = source


# Not quite sure what customization needed yet
class ChessDBrecordRepertoire(ChessDBrecordGame):

    """Repertoire record customised for exporting repertoire information."""

    def __init__(self):
        """Extend with move number encode and decode methods"""
        # Skip the immediate superclass __init__ to fix key and value classes
        super(ChessDBrecordGame, self).__init__(
            ChessDBkeyGame, ChessDBvalueRepertoire
        )


# Not quite sure what customization needed yet
class ChessDBrecordRepertoireTags(ChessDBrecordGameTags):

    """Repertoire record customised for displaying repertoire tag information."""

    def __init__(self):
        """Extend with move number encode and decode methods"""
        # Skip the immediate superclass __init__ to fix key and value classes
        super(ChessDBrecordGameTags, self).__init__(
            ChessDBkeyGame, ChessDBvalueRepertoireTags
        )


# Not quite sure what customization needed yet
class ChessDBrecordRepertoireUpdate(ChessDBrecordGameUpdate):

    """Repertoire record customized for editing repertoire records."""

    def __init__(self):
        """Extend with move number encode and decode methods"""
        # Skip the immediate superclass __init__ to fix key and value classes
        super(ChessDBrecordGameUpdate, self).__init__(
            ChessDBkeyGame, ChessDBvalueRepertoireUpdate
        )

    def get_keys(self, datasource=None, partial=None):
        """Return list of (key, value) tuples.

        The keys for the secondary databases in a ChessDatabase instance are
        embedded in, or derived from, the PGN string for the game.  All
        except the positions are held in self.value.collected_game._tags.
        Multiple position keys can arise becuse all repetitions of a
        position are of interest.  The partial argument is used to select
        the relevant keys.  The position is in partial and the keys will
        differ in the move number.

        """
        dbname = datasource.dbname
        if dbname == REPERTOIRE_FILE_DEF:
            return [(self.key.recno, self.srvalue)]
        elif dbname in self.value.collected_game._tags:
            return [(self.value.collected_game._tags[dbname], self.key.pack())]
        else:
            return []


class ChessDBvalueAnalysis(Analysis, Value):

    """Chess engine analysis data for a position."""

    def __init__(self):
        """ """
        super().__init__()

    def pack(self):
        """Extend, return analysis record and index data."""
        v = super().pack()
        index = v[1]
        index[VARIATION_FIELD_DEF] = [self.position]
        index[ENGINE_FIELD_DEF] = [k for k in self.scale]
        return v


class ChessDBrecordAnalysis(Record):

    """Chess game record customised for chess engine analysis data.

    No index values are derived from PGN move text, so there is no advantage in
    separate classes for display and update.  The PGN FEN tag provides the only
    PGN related index value used.

    """

    def __init__(self):
        """Delegate using ChessDBkeyGame and ChessDBvalueAnalysis classes."""
        super().__init__(KeyData, ChessDBvalueAnalysis)


class ChessDBkeyQuery(KeyData):

    """Primary key of game selection rule record."""


class ChessDBvalueQuery(QueryStatement, Value):

    """Game selection rule data."""

    def __init__(self):
        super(ChessDBvalueQuery, self).__init__()

    def __eq__(self, other):
        """Return (self == other).  Attributes are compared explicitly."""
        try:
            if (
                self.get_name_query_statement_text()
                != other.get_name_query_statement_text()
            ):
                return False
            else:
                return True
        except:
            return False

    def __ne__(self, other):
        """Return (self != other).  Attributes are compared explicitly."""
        try:
            if (
                self.get_name_query_statement_text()
                == other.get_name_query_statement_text()
            ):
                return False
            else:
                return True
        except:
            return True

    def load(self, value):
        """Set game selection rule from value"""
        self.process_query_statement(literal_eval(value))

    def pack_value(self):
        """Return gameselection rule value"""
        return repr(self.get_name_query_statement_text())

    def pack(self):
        """Extend, return game selection rule record and index data."""
        v = super().pack()
        index = v[1]
        index[RULE_FIELD_DEF] = [self.get_name_text()]
        return v


class ChessDBrecordQuery(Record):

    """Game selection rule record."""

    def __init__(self):
        """Extend as a game selection rule record."""
        super(ChessDBrecordQuery, self).__init__(
            ChessDBkeyQuery, ChessDBvalueQuery
        )

    def get_keys(self, datasource=None, partial=None):
        """Return list of (key, value) tuples.

        The game selection rule name is held in an attribute which is not named
        for the field where it exists in the database.

        """
        if datasource.dbname == RULE_FIELD_DEF:
            return [(self.value.get_name_text(), self.key.pack())]
        else:
            return super().get_keys(datasource=datasource, partial=partial)

    def load_value(self, value):
        """Load self.value from value which is repr(<data>).

        Set database in self.value for query processing then delegate value
        processing to superclass.

        """
        self.value.set_database(self.database)
        self.value.dbset = GAMES_FILE_DEF
        super().load_value(value)


class ChessDBkeyEngine(KeyData):

    """Primary key of chess engine record."""


class ChessDBvalueEngine(Engine, Value):

    """Game selection rule data."""

    def __init__(self):
        super(ChessDBvalueEngine, self).__init__()

    def pack(self):
        """Extend, return game selection rule record and index data."""
        v = super().pack()
        index = v[1]
        index[COMMAND_FIELD_DEF] = [self.get_name_text()]
        return v


class ChessDBrecordEngine(Record):

    """Chess engine record."""

    def __init__(self):
        """Extend as a game selection rule record."""
        super(ChessDBrecordEngine, self).__init__(
            ChessDBkeyEngine, ChessDBvalueEngine
        )



# blanktext.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Define tkinter.Text widget to be customized and used within ChessTab.

The positionrow.PositionRow and positionscore.PositionScore classes do not use
this class because their Text widgets come from solentware_grid.gui.DataRow.

"""

import tkinter
import enum

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from .eventspec import EventSpec
from .displayitems import DisplayItemsStub


# 'Tag' in these names refers to tags in Tk Text widgets, not PGN tags.
# NO_EDITABLE_TAGS and INITIAL_BINDINGS are used if _is_text_editable is False.
# All subclasses use DEFAULT_BINDINGS is _is_text_editable is True.
# Score uses NO_CURRENT_TOKEN, CURRENT_NO_TAGS, and SELECT_VARIATION, which
# other subclasses do not.
# GameEdit uses all the names.
class NonTagBind(enum.Enum):
    # Current token is a move.
    NO_EDITABLE_TAGS = 1
    NO_CURRENT_TOKEN = 2
    DEFAULT_BINDINGS = 3
    INITIAL_BINDINGS = 4
    CURRENT_NO_TAGS = 5
    # Current token is a move with variations for next move, and the attempt
    # to go to next move was intercepted to choose which one.
    SELECT_VARIATION = 6


class BlankText(ExceptionHandler):

    """Create Text widget with configuration shared by subclasses.

    The subclasses are cqltext.CQLText, querytext.QueryText, score.Score,
    and enginetext.EngineText.

    panel is used as the master argument for the tkinter Text and Scrollbar
    widgets created to display the statement text.

    items_manager is the ui attribute which tracks which EngineText instance is
    active (as defined by ui).

    Subclasses are responsible for providing a geometry manager.

    Attribute _is_text_editable is False to indicate text cannot be edited.

    Attribute _most_recent_bindings is set to indicate the initial set of
    event bindings.  Instances will override this as required.

    """

    # True means the content can be edited.
    _is_text_editable = False

    # Indicate the most recent set of bindings applied to score attribute.
    # Values are Tk tag names or members of NonTagBind enumeration.
    _most_recent_bindings = NonTagBind.INITIAL_BINDINGS

    def __init__(self, panel, items_manager=None, **ka):
        """Create widgets to display chess engine definition."""
        super().__init__(**ka)

        # May be worth using a Null() instance for these two attributes.
        if items_manager is None:
            items_manager = DisplayItemsStub()
        self.items = items_manager

        self.panel = panel
        self.score = tkinter.Text(
            master=self.panel,
            width=0,
            height=0,
            takefocus=tkinter.FALSE,
            undo=True,
            wrap=tkinter.WORD,
        )
        self.scrollbar = tkinter.Scrollbar(
            master=self.panel,
            orient=tkinter.VERTICAL,
            takefocus=tkinter.FALSE,
            command=self.score.yview,
        )
        self.score.configure(yscrollcommand=self.scrollbar.set)

        # Keyboard actions do nothing by default.
        self.set_keypress_binding(switch=False)
        self.set_event_bindings_score(self.get_menubar_events())

        # The popup menus used by all subclasses.
        self.inactive_popup = None

    def set_event_bindings_score(self, bindings=(), switch=True):
        """Set bindings if switch is True or unset the bindings."""
        ste = self.try_event
        for sequence, function in bindings:
            self.score.bind(
                sequence[0], ste(function) if switch and function else ""
            )

    def set_keypress_binding(self, function=None, bindings=(), switch=True):
        """Set bindings to function if switch is True or disable keypress."""
        if switch and function:
            stef = self.try_event(function)
            for sequence in bindings:
                self.score.bind(sequence[0], stef)
        else:
            stekb = self.try_event(self.press_break)
            for sequence in bindings:
                self.score.bind(sequence[0], stekb)

    def get_menubar_events(self):
        """Return tuple of event binding definitions passed for menubar."""
        return ((EventSpec.score_enable_F10_menubar, self.press_none),)

    def press_break(self, event=None):
        """Do nothing and prevent event handling by next handlers."""
        return "break"

    def press_none(self, event=None):
        """Do nothing and allow event to be handled by next handler."""
        return None

    # This method arose when seeking clarity in the way popup menus were set,
    # and replaces lots of 'add_command' calls scattered all over.
    # Long term, either this method or add_cascade_menu_to_popup will do all.
    def set_popup_bindings(self, popup, bindings=(), index=tkinter.END):
        """Insert bindings in popup before index in popup."""

        # Default index is tkinter.END which seems to mean insert at end of
        # popup, not before last entry in popup as might be expected from the
        # way expressed in the 'Tk menu manual page' for index command.  (The
        # manual page describes 'end' in the context of 'none' for 'activate'
        # option.  It does make sense 'end' meaning after existing entries
        # when inserting entries.)

        for accelerator, function in bindings:
            popup.insert_command(
                index=index,
                label=accelerator[1],
                command=self.try_command(function, popup),
                accelerator=accelerator[2],
            )

    def give_focus_to_widget(self, event=None):
        """Do nothing and return 'break'.  Override in subclasses as needed."""
        return "break"

    def get_F10_popup_events(self, top_left, pointer):
        """Return tuple of event definitions to post popup menus at top left
        of focus widget and at pointer location within application widget.

        top_left and pointer are functions.

        """
        return (
            (EventSpec.score_enable_F10_popupmenu_at_top_left, top_left),
            (EventSpec.score_enable_F10_popupmenu_at_pointer, pointer),
        )

    # Subclasses with database interfaces may override method.
    def create_database_submenu(self, menu):
        return None

    def post_menu(self, menu, create_menu, allowed=True, event=None):
        if menu is None:
            menu = create_menu()
        if not allowed:
            return "break"
        menu.tk_popup(*self.score.winfo_pointerxy())

        # So 'Control-F10' does not fire 'F10' (menubar) binding too.
        return "break"

    def post_menu_at_top_left(
        self, menu, create_menu, allowed=True, event=None
    ):
        if menu is None:
            menu = create_menu()
        if not allowed:
            return "break"
        menu.tk_popup(event.x_root - event.x, event.y_root - event.y)

        # So 'Shift-F10' does not fire 'F10' (menubar) binding too.
        return "break"

    def is_active_item_mapped(self):
        """ """
        if self.items.is_mapped_panel(self.panel):
            if self is not self.items.active_item:
                return False
        return True

    def bind_for_primary_activity(self, switch=True):
        """Set (switch True) or clear bindings for main actions when active.

        If bool(switch) is true, clear the most recently set bindings first.

        """
        if switch:
            self.token_bind_method[self._most_recent_bindings](self, False)
            self._most_recent_bindings = NonTagBind.NO_EDITABLE_TAGS
        self.set_primary_activity_bindings(switch=switch)

    def bind_for_initial_state(self, switch=True):
        """Clear the most recently set bindings if bool(switch) is True.

        Assume not setting new bindings leaves widget in initial state.

        If bool(switch) is False, nothing is done.

        """
        if switch:
            self.token_bind_method[self._most_recent_bindings](self, False)
            self._most_recent_bindings = NonTagBind.INITIAL_BINDINGS

    # Dispatch dictionary for token binding selection.
    # Keys are the possible values of self._most_recent_bindings.
    token_bind_method = {
        NonTagBind.NO_EDITABLE_TAGS: bind_for_primary_activity,
        NonTagBind.INITIAL_BINDINGS: bind_for_initial_state,
    }

    def create_primary_activity_popup(self):
        assert self.primary_activity_popup is None
        popup = tkinter.Menu(master=self.score, tearoff=False)
        self.set_popup_bindings(popup, self.get_primary_activity_events())
        database_submenu = self.create_database_submenu(popup)
        if database_submenu:
            popup.add_cascade(label="Database", menu=database_submenu)
        self.primary_activity_popup = popup
        return popup

    def get_button_events(self, buttonpress1=None, buttonpress3=None):
        """Return tuple of buttonpress event bindings.

        buttonpress1 and buttonpress3 default to self.press_none().

        """
        if buttonpress1 is None:
            buttonpress1 = self.press_none
        if buttonpress3 is None:
            buttonpress3 = self.press_none
        return self.get_modifier_buttonpress_suppression_events() + (
            (EventSpec.buttonpress_1, buttonpress1),
            (EventSpec.buttonpress_3, buttonpress3),
        )

    # Take a snapshot of the tkinter.Text widget bound to self.score.  It is
    # intended for problem tracing.  It is saved as a sibling of ErrorLog.
    # At a convenient time dump_text_widget will become a method in the
    # solentware_misc.gui.exceptionhandler.ExceptionHandler class.
    def dump_text_widget(self, textwidget, filename=None):
        if filename is None:
            filename = "dumptextwidget"
        import os
        import datetime

        filename = "_".join(
            (
                filename,
                datetime.datetime.now(datetime.timezone.utc).isoformat(),
            )
        )
        with open(
            os.path.join(
                os.path.dirname(self.get_error_file_name()), filename
            ),
            "w",
        ) as f:
            for t in textwidget.dump("1.0", tkinter.END):
                f.write(repr(t) + "\n")



# __init__.py
# Copyright 2020 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a NoSQL database of chess games using the dbm.ndbm package.

This interface can be used if ndbm software is installed and available to the
dbm.ndbm package.
"""



# cqltoplevel.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Toplevel widgets to display and edit Chess Query Language (ChessQL)
statements.

ChessQL statements obey the syntax published for CQL version 6.0.1 (by Gady
Costeff).

These two classes display ChessQL statements in their own Toplevel widget:
they are used in the cqldbdelete, cqldbedit, and cqldbshow, modules.

"""

from .cql import CQL
from .cqledit import CQLEdit
from .topleveltext import ToplevelText


class CQLToplevel(ToplevelText, CQL):

    """Customize CQL to be the single instance in a Toplevel widget."""


class CQLToplevelEdit(ToplevelText, CQLEdit):

    """Customize CQLEdit to be the single instance in a Toplevel widget."""

    # A method like GameToplevelEdit.create_primary_activity_popup is not
    # needed because the standard edit operations of Text widget are
    # sufficient.



# enginedbdelete.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise delete toplevel to delete chess engine record.
"""

from solentware_grid.gui.datadelete import DataDelete

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from .enginetoplevel import EngineToplevel
from .topleveltext import DeleteText


class EngineDbDelete(ExceptionHandler, DeleteText, DataDelete):
    """Delete a chess engine definition from database.

    parent is used as the master argument in a EngineToplevel call.

    ui is used as the ui argument in a EngineToplevel call.

    parent, oldobject, and the EngineToplevel instance created, are used as
    arguments in the super.__init__ call.

    Attribute text_name provides the name used in widget titles and message
    text.

    Methods get_title_for_object and set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    text_name = "Engine Definition"

    def __init__(self, parent, oldobject, ui=None):
        """Extend and create toplevel widget to delete chess engine definition.

        ui should be a UCI instance.

        """
        # Toplevel title set '' in __init__ and to proper value in initialize.
        super().__init__(
            oldobject, parent, EngineToplevel(master=parent, ui=ui), ""
        )
        self.initialize()

    def get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a chess engine definition
        object_.

        Default value of object_ is object attribute from DataDelete class.

        """
        if object_ is None:
            object_ = self.object
        return "  ".join(
            (
                self.text_name.join(("Delete ", ":")),
                object_.value.get_name_text(),
            )
        )

    @property
    def ui_base_table(self):
        return self.ui.base_engines

    @property
    def ui_items_in_toplevels(self):
        return self.ui.engines_in_toplevels

    @property
    def ui(self):
        return self.oldview.ui

    def set_item(self, view, object_):
        view.definition.extract_engine_definition(object_.get_srvalue())
        view.set_engine_definition(object_.value)

    def tidy_on_destroy(self):
        # ui_base_table is None when this happens other than directly closing
        # the Toplevel.
        try:
            super().tidy_on_destroy()
        except AttributeError:
            if self.ui_base_table is not None:
                raise



# querydisplay.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widgets to display and edit game selection rules.

These four classes display game selection rules in the main window: they are
used in the querygrid module.

The _QueryDisplay class provides attributes and behaviour shared by the
QueryDisplay, QueryDisplayInsert, and QueryDisplayEdit, classes.  It also
provides properties to support implementation of behaviour shared with the
Game*, Repertoire*, and CQL*, classes.

The QueryDisplay, QueryDisplayInsert, and QueryDisplayEdit, classes are
subclasses of the relevant ShowText, InsertText, EditText, DisplayText, and
ListGamesText, classes from the displaytext module; to implement behaviour
shared with all Text widgets in the main display (that includes widgets
displaying PGN text).

"""

import tkinter
import tkinter.messagebox

from solentware_base.core.where import WhereError

from solentware_grid.gui.dataedit import RecordEdit
from solentware_grid.gui.datadelete import RecordDelete
from solentware_grid.core.dataclient import DataNotify

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from ..core.querystatement import QueryStatement
from .query import Query
from .queryedit import QueryEdit
from ..core.chessrecord import ChessDBrecordQuery
from .eventspec import EventSpec
from .display import Display
from .displaytext import (
    ShowText,
    InsertText,
    EditText,
    DisplayText,
    ListGamesText,
)


class _QueryDisplay(ExceptionHandler, Display):

    """Extend and link game selection rule to database.

    sourceobject - link to database.

    Attribute binding_labels specifies the order navigation bindings appear
    in popup menus.

    Method insert_item_database allows records to be inserted into a database
    from any CQL widget.

    """

    binding_labels = (
        EventSpec.navigate_to_position_grid,
        EventSpec.navigate_to_active_game,
        EventSpec.navigate_to_game_grid,
        EventSpec.navigate_to_repertoire_grid,
        EventSpec.navigate_to_active_repertoire,
        EventSpec.navigate_to_repertoire_game_grid,
        EventSpec.navigate_to_partial_grid,
        EventSpec.navigate_to_active_partial,
        EventSpec.navigate_to_selection_rule_grid,
        EventSpec.selectiondisplay_to_previous_selection,
        EventSpec.selectiondisplay_to_next_selection,
        EventSpec.navigate_to_partial_game_grid,
        EventSpec.tab_traverse_backward,
        EventSpec.tab_traverse_forward,
    )

    def __init__(self, sourceobject=None, **ka):
        """Extend and link game selection rule to database."""
        super().__init__(**ka)
        self.blockchange = False
        if self.ui.base_selections.datasource:
            self.set_data_source(self.ui.base_selections.get_data_source())
        self.sourceobject = sourceobject
        self.insertonly = sourceobject is None
        self.recalculate_after_edit = sourceobject

    @property
    def ui_displayed_items(self):
        return self.ui.selection_items

    @property
    def ui_configure_item_list_grid(self):
        return self.ui.configure_selection_grid

    @property
    def ui_set_item_name(self):
        return self.ui._set_selection_name

    @property
    def ui_set_find_item_games(self):
        return self.ui._set_find_selection_name_games

    def get_navigation_events(self):
        """Return event description tuple for navigation from selection rule
        statement."""
        return (
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (EventSpec.navigate_to_active_game, self.set_focus_gamepanel_item),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_active_repertoire,
                self.set_focus_repertoirepanel_item,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_selection_rule,
                self.set_focus_selectionpanel_item,
            ),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.selectiondisplay_to_previous_selection,
                self.prior_item,
            ),
            (EventSpec.selectiondisplay_to_next_selection, self.next_item),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_round, self.traverse_round),
        )

    def delete_item_view(self, event=None):
        """Remove game selection rule item from screen."""
        self.ui.delete_selection_rule_view(self)

    def insert_item_database(self, event=None):
        """Add game selection rule to database."""
        if self.ui.selection_items.active_item is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Insert Game Selection Rule",
                message="No active game selection rule to insert into database.",
            )
            return

        # This should see if game selection rule with same name already exists,
        # after checking for database open, and offer option to insert anyway.
        if self.ui.database is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Insert Game Selection Rule",
                message="Cannot add game selection rule:\n\nNo database open.",
            )
            return

        datasource = self.ui.base_selections.get_data_source()
        if datasource is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Insert Game Selection Rule",
                message="Cannot add game selection rule:\n\nRule list hidden.",
            )
            return
        updater = ChessDBrecordQuery()
        updater.value.set_database(datasource.dbhome)
        updater.value.dbset = self.ui.base_games.datasource.dbset
        uvpqs = updater.value.process_query_statement(
            self.get_name_query_statement_text()
        )
        if not len(updater.value.get_name_text()):
            tkinter.messagebox.showerror(
                parent=self.ui.get_toplevel(),
                title="Insert Game Selection Rule",
                message="".join(
                    (
                        "The selection rule has no name.\n\nPlease enter it's ",
                        "name as the first line of text.'",
                    )
                ),
            )
            return
        if not uvpqs:
            if not updater.value.where_error:
                if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                    parent=self.ui.get_toplevel(),
                    title="Insert Game Selection Rule",
                    message="".join(
                        (
                            "Confirm request to add game selection rule named:\n\n",
                            updater.value.get_name_text(),
                            "\n\nto database.\n\n",
                            "Note validation of the statement failed but no ",
                            "information is available.",
                        )
                    ),
                ):
                    tkinter.messagebox.showinfo(
                        parent=self.ui.get_toplevel(),
                        title="Insert Game Selection Rule",
                        message="".join(
                            (
                                "Add game selection rule to ",
                                "database abandonned.",
                            )
                        ),
                    )
                    return
            elif tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                parent=self.ui.get_toplevel(),
                title="Insert Game Selection Rule",
                message="".join(
                    (
                        "Confirm request to add game selection rule named:\n\n",
                        updater.value.get_name_text(),
                        "\n\nto database.\n\n",
                        updater.value.where_error.get_error_report(
                            self.ui.base_games.get_data_source()
                        ),
                    )
                ),
            ):
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="Insert Game Selection Rule",
                    message="".join(
                        ("Add game selection rule to ", "database abandonned.")
                    ),
                )
                return
        elif tkinter.messagebox.YES != tkinter.messagebox.askquestion(
            parent=self.ui.get_toplevel(),
            title="Insert Game Selection Rule",
            message="".join(
                (
                    "Confirm request to add game selection rule named:\n\n",
                    updater.value.get_name_text(),
                    "\n\nto database.",
                )
            ),
        ):
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Insert Game Selection Rule",
                message="Add game selection rule to database abandonned.",
            )
            return
        editor = RecordEdit(updater, None)
        editor.set_data_source(datasource, editor.on_data_change)
        updater.set_database(editor.get_data_source().dbhome)
        updater.key.recno = None  # 0
        editor.put()
        tkinter.messagebox.showinfo(
            parent=self.ui.get_toplevel(),
            title="Insert Game Selection Rule",
            message="".join(
                (
                    'Game selection rule "',
                    updater.value.get_name_text(),
                    '" added to database.',
                )
            ),
        )

    # on_game_change() is not called, but do nothing if it is.
    def on_game_change(self, instance):
        """Prevent update from self if instance refers to same record."""
        if self.sourceobject is not None:
            pass

    def get_list_games_events(self):
        """Return tuple of event bindings to list games for selection rule."""
        return (
            (EventSpec.display_list, self.process_and_set_selection_rule_list),
        )

    def set_database_navigation_close_item_bindings(self, switch=True):
        super().set_database_navigation_close_item_bindings(switch=switch)
        self.set_event_bindings_score(
            self.get_list_games_events(), switch=switch
        )

    def add_list_games_entry_to_popup(self, popup, index=tkinter.END):
        """Add option to list games for selection rule to popup."""
        self.set_popup_bindings(
            popup, self.get_list_games_events(), index=index
        )

    def generate_popup_navigation_maps(self):
        navigation_map = {k: v for k, v in self.get_navigation_events()}
        local_map = {}
        return navigation_map, local_map

    def create_primary_activity_popup(self):
        popup = super().create_primary_activity_popup()
        self.add_close_item_entry_to_popup(popup)
        return popup

    def on_selection_change(self, instance):
        """Prevent update from self if instance refers to same record.

        The list of games is updated by the 'List Games' popup menu option,
        but not as a consequence of updating the database.
        """
        if instance.newrecord:

            # Editing an existing record.
            value = instance.newrecord.value
            key = instance.newrecord.key

        else:

            # Inserting a new record.
            value = instance.value
            key = instance.key

        if self.sourceobject is not None:
            if (
                key == self.sourceobject.key
                and self.datasource.dbname == self.sourceobject.dbname
                and self.datasource.dbset == self.sourceobject.dbset
            ):
                self.blockchange = True

    def get_text_for_statusbar(self):
        """ """
        return "".join(
            (
                "Please wait while finding games for game selection rule ",
                self.query_statement.get_name_text(),
            )
        )

    def get_selection_text_for_statusbar(self):
        """ """
        return self.query_statement.get_name_text()

    # This method added so it can be used in _cycle_items which then becomes
    # identical to definitions in other classes.  There may be a case for
    # following what cqldisplay does in it's set_game_list method.
    def set_game_list(self):
        """Delegate to refresh_game_list.

        Enables 'after(refresh_game_list, ...)' in give_focus_to_widget()
        without affecting set_game_list() call in _cycle_items method.

        """
        self.refresh_game_list()

    def process_and_set_selection_rule_list(self, event=None):
        """Display games matching edited game selection rule."""
        if self.ui.base_games.datasource:
            self.query_statement.set_database(
                self.ui.base_games.datasource.dbhome
            )
        try:
            self.query_statement.process_query_statement(
                self.score.get("1.0", tkinter.END)
            )
        except WhereError as exc:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=" ".join(("Game Selection Rule Error")),
                message=str(exc),
            )
            return "break"
        finally:
            self.query_statement.set_database()
        self.refresh_game_list()
        return "break"


class QueryDisplay(_QueryDisplay, DisplayText, ShowText, Query, DataNotify):

    """Display game selection rule from database and allow delete and insert.

    Method delete_item_database allows records to be deleted from a database.

    """

    # QueryDisplay has this method, but CQLDisplay does not, because the game
    # list area is shared with the main game list and the index lists.
    # Generating the game list on demand may be necessary at any time.
    def create_primary_activity_popup(self):
        popup = super().create_primary_activity_popup()
        self.add_list_games_entry_to_popup(popup, index="Close Item")
        return popup

    def delete_item_database(self, event=None):
        """Remove game selection rule from database."""
        if self.ui.database is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Delete Game Selection Rule",
                message="".join(
                    (
                        "Cannot delete game selection rule:\n\n",
                        "No database open.",
                    )
                ),
            )
            return
        datasource = self.ui.base_selections.get_data_source()
        if datasource is None:
            tkinter.messagebox.showerror(
                parent=self.ui.get_toplevel(),
                title="Delete Game Selection Rule",
                message="".join(
                    (
                        "Cannot delete game selection rule:\n\n",
                        "Rule list hidden.",
                    )
                ),
            )
            return
        if self.sourceobject is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Delete Game Selection Rule",
                message="".join(
                    (
                        "The game selection rule to delete has not ",
                        "been given.\n\nProbably because database ",
                        "has been closed and opened since this copy ",
                        "was displayed.",
                    )
                ),
            )
            return
        if self.blockchange:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Delete Game Selection Rule",
                message="\n".join(
                    (
                        "Cannot delete game selection rule.",
                        "Record has been amended since this copy displayed.",
                    )
                ),
            )
            return
        if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
            parent=self.ui.get_toplevel(),
            title="Delete Game Selection Rule",
            message="Confirm request to delete game selection rule.",
        ):
            return
        s = self.query_statement
        if s.where_error is not None:
            v = self.sourceobject.value
            if (
                s.get_name_text() != v.get_name_text()
                or s.where_error != v.where_error
                or s.get_query_statement_text() != v.get_query_statement_text()
            ):
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="Delete Game Selection Rule",
                    message="\n".join(
                        (
                            "Cannot delete game selection rule.",
                            " ".join(
                                (
                                    "Rule on display is not same as",
                                    "rule from record.",
                                )
                            ),
                        )
                    ),
                )
                return
        editor = RecordDelete(self.sourceobject)
        editor.set_data_source(datasource, editor.on_data_change)
        editor.delete()
        tkinter.messagebox.showinfo(
            parent=self.ui.get_toplevel(),
            title="Delete Game Selection Rule",
            message="".join(
                (
                    'Game selection rule "',
                    self.sourceobject.value.get_name_text(),
                    '" deleted from database.',
                )
            ),
        )


class QueryDisplayInsert(
    _QueryDisplay, ListGamesText, InsertText, ShowText, QueryEdit, DataNotify
):

    """Display game selection rule from database allowing insert.

    QueryEdit provides the widget and _QueryDisplay the database interface.

    Listing games for ChessQL statements is different to selection rule
    statements because selection rules share the listing area with the Game
    and Index options to the Select menu.  Index opens up the 'Move to' and
    Filter options too.  A definite user action is always required to generate
    game lists for selection rules, in addition to navigating to (giving focus
    to) the selection rule.  The area where ChessQL game lists are shown is
    dedicated to ChessQL statements.

    """


class QueryDisplayEdit(EditText, QueryDisplayInsert):

    """Display game selection rule from database allowing edit and insert.

    Method update_item_database allows records on the database to be amended.

    """

    def update_item_database(self, event=None):
        """Modify existing game selection rule record."""
        if self.ui.database is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Edit Game Selection Rule",
                message="Cannot edit repertoire:\n\nNo database open.",
            )
            return
        datasource = self.ui.base_selections.get_data_source()
        if datasource is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Edit Game Selection Rule",
                message="Cannot edit game selection rule:\n\nRule list hidden.",
            )
            return
        if self.sourceobject is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Edit Game Selection Rule",
                message="".join(
                    (
                        "The game selection rule to edit has not ",
                        "been given.\n\nProbably because database ",
                        "has been closed and opened since this copy ",
                        "was displayed.",
                    )
                ),
            )
            return
        if self.blockchange:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Edit Game Selection Rule",
                message="\n".join(
                    (
                        "Cannot edit game selection rule.",
                        "Rule has been amended since this copy was displayed.",
                    )
                ),
            )
            return
        original = ChessDBrecordQuery()
        original.set_database(datasource.dbhome)
        original.load_record(
            (self.sourceobject.key.recno, self.sourceobject.srvalue)
        )

        # is it better to use DataClient directly?
        # Then original would not be used. Instead DataSource.new_row
        # gets record keyed by sourceobject and update is used to edit this.
        updater = ChessDBrecordQuery()
        updater.value.set_database(datasource.dbhome)
        updater.value.dbset = self.ui.base_games.datasource.dbset
        uvpqs = updater.value.process_query_statement(
            self.get_name_query_statement_text()
        )
        if not len(updater.value.get_name_text()):
            tkinter.messagebox.showerror(
                parent=self.ui.get_toplevel(),
                title="Edit Game Selection Rule",
                message="".join(
                    (
                        "The selection rule has no name.\n\nPlease enter it's ",
                        "name as the first line of text.'",
                    )
                ),
            )
            return
        if not uvpqs:
            if not updater.value.where_error:
                if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                    parent=self.ui.get_toplevel(),
                    title="Edit Game Selection Rule",
                    message="".join(
                        (
                            "Confirm request to edit game selection rule ",
                            "named:\n\n",
                            updater.value.get_name_text(),
                            "\n\nto database.\n\n",
                            "Note validation of the statement failed but no ",
                            "information is available.",
                        )
                    ),
                ):
                    tkinter.messagebox.showinfo(
                        parent=self.ui.get_toplevel(),
                        title="Edit Game Selection Rule",
                        message="".join(
                            (
                                "Edit game selection rule to ",
                                "database abandonned.",
                            )
                        ),
                    )
                    return
            elif tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                parent=self.ui.get_toplevel(),
                title="Edit Game Selection Rule",
                message="".join(
                    (
                        "Confirm request to edit game selection rule named:\n\n",
                        updater.value.get_name_text(),
                        "\n\nto database.\n\n",
                        updater.value.where_error.get_error_report(
                            self.ui.base_games.get_data_source()
                        ),
                    )
                ),
            ):
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="Edit Game Selection Rule",
                    message="".join(
                        (
                            "Edit game selection rule to ",
                            "database abandonned.",
                        )
                    ),
                )
                return
        elif tkinter.messagebox.YES != tkinter.messagebox.askquestion(
            parent=self.ui.get_toplevel(),
            title="Edit Game Selection Rule",
            message="".join(
                (
                    "Confirm request to edit game selection rule named:\n\n",
                    updater.value.get_name_text(),
                    "\n\non database.",
                )
            ),
        ):
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="Edit Game Selection Rule",
                message="Edit game selection rule to database abandonned.",
            )
            return
        editor = RecordEdit(updater, original)
        editor.set_data_source(datasource, editor.on_data_change)
        updater.set_database(editor.get_data_source().dbhome)
        original.set_database(editor.get_data_source().dbhome)
        updater.key.recno = original.key.recno
        editor.edit()
        tkinter.messagebox.showinfo(
            parent=self.ui.get_toplevel(),
            title="Edit Game Selection Rule",
            message="".join(
                (
                    'Game selection rule "',
                    updater.value.get_name_text(),
                    '" amended on database.',
                )
            ),
        )



# repertoiredisplay.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widgets to display and edit repertoires.

These four classes display PGN text for games in the main window: they are
used in the gamelistgrid module.

The _RepertoireDisplay class provides attributes and behaviour shared by the
RepertoireDisplay, RepertoireDisplayInsert, and RepertoireDisplayEdit,
classes.  It also provides properties to support implementation of behaviour
shared with the CQL*, Game*, and Query*, classes.

The RepertoireDisplay, RepertoireDisplayInsert, and RepertoireDisplayEdit,
classes are subclasses of the relevant ShowPGN, InsertPGN, EditPGN, and
DisplayPGN, classes from the displaypgn module; to implement behaviour shared
with all text widgets in the main display (that includes widgets displaying
text).

"""

import tkinter
import tkinter.messagebox

from solentware_grid.core.dataclient import DataNotify

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from ..core.constants import TAG_OPENING
from .repertoire import Repertoire
from .repertoireedit import RepertoireEdit
from ..core.chessrecord import ChessDBrecordRepertoireUpdate
from .eventspec import EventSpec
from .display import Display
from .displaypgn import ShowPGN, InsertPGN, EditPGN, DisplayPGN


class _RepertoireDisplay(ExceptionHandler, Display):

    """Extend and link PGN repertoire text to database.

    sourceobject - link to database.

    Attribute binding_labels specifies the order navigation bindings appear
    in popup menus.

    Attribute pgn_score_name provides the name used in widget titles and
    message text.

    Attribute pgn_score_tags provides the PGN tag names used in widget titles
    and message text.  It is the opening PGN tag defined in ChessTab.

    Attribute pgn_score_source_name provides the error key value to index a
    PGN game score with errors.

    Attribute pgn_score_updater provides the class used to process PGN text
    into a database update.

    """

    binding_labels = (
        EventSpec.navigate_to_position_grid,
        EventSpec.navigate_to_active_game,
        EventSpec.navigate_to_game_grid,
        EventSpec.navigate_to_repertoire_grid,
        EventSpec.repertoiredisplay_to_previous_repertoire,
        EventSpec.analysis_to_scoresheet,
        EventSpec.repertoiredisplay_to_next_repertoire,
        EventSpec.navigate_to_repertoire_game_grid,
        EventSpec.scoresheet_to_analysis,
        EventSpec.navigate_to_partial_grid,
        EventSpec.navigate_to_active_partial,
        EventSpec.navigate_to_partial_game_grid,
        EventSpec.navigate_to_selection_rule_grid,
        EventSpec.navigate_to_active_selection_rule,
        EventSpec.tab_traverse_backward,
        EventSpec.tab_traverse_forward,
    )

    # These exist so the insert_game_database methods in _RepertoireDisplay and
    # gamedisplay._GameDisplay, and delete_game_database in RepertoireDisplay
    # and gameedisplay.GameDisplay, can be modified and replaced by single
    # copies in the displaypgn.ShowPGN class.
    # See mark_partial_positions_to_be_recalculated() method too.
    # The names need to be more generic to make sense in cql, engine, and
    # query, context.
    pgn_score_name = "repertoire"
    pgn_score_source_name = "No opening name"
    pgn_score_tags = (TAG_OPENING,)
    pgn_score_updater = ChessDBrecordRepertoireUpdate

    def __init__(self, sourceobject=None, **ka):
        """Extend and link repertoire to database."""
        super().__init__(**ka)
        self.blockchange = False
        if self.ui.base_repertoires.datasource:
            self.set_data_source(self.ui.base_repertoires.get_data_source())
        self.sourceobject = sourceobject

    # Could be put in game.Repertoire class to hide the game.Game version.
    # Here can be justified because purpose is allow some methods to be moved
    # to displaypgn.ShowPGN class.
    @property
    def ui_displayed_items(self):
        return self.ui.repertoire_items

    # Defined so cycle_item and give_focus_to_widget methods can be shared by
    # gamedisplay._GameDisplay and repertoiredisplay._RepertoireDisplay classes.
    @property
    def ui_configure_item_list_grid(self):
        return self.ui.configure_repertoire_grid

    # ui_base_table and mark_partial_positions_to_be_recalculated defined
    # so insert_game_database method can be shared by gamedisplay._GameDisplay
    # and repertoiredisplay._RepertoireDisplay classes.
    # See class attributes pgn_score_name and pgn_score_source_name too.

    @property
    def ui_base_table(self):
        return self.ui.base_repertoires

    @property
    def ui_items_in_toplevels(self):
        return self.ui.games_and_repertoires_in_toplevels

    def mark_partial_positions_to_be_recalculated(self, datasource=None):
        pass

    def get_navigation_events(self):
        """Return event description tuple for navigation from repertoire."""
        return (
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_partial,
                self.set_focus_partialpanel_item,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (EventSpec.navigate_to_active_game, self.set_focus_gamepanel_item),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self.set_focus_selectionpanel_item,
            ),
            (
                EventSpec.repertoiredisplay_to_previous_repertoire,
                self.prior_item,
            ),
            (EventSpec.repertoiredisplay_to_next_repertoire, self.next_item),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            # No traverse_round because Alt-F8 toggles repertoire and analysis.
        )

    def delete_item_view(self, event=None):
        """Remove repertoire item from screen."""
        self.set_data_source()
        self.ui.delete_repertoire_view(self)

    def on_game_change(self, instance):
        """Prevent update from self if instance refers to same record."""
        if self.sourceobject is not None:
            self.patch_pgn_score_to_fit_record_change_and_refresh_grid(
                self.ui.repertoire_games, instance
            )

    def on_repertoire_change(self, instance):
        """Prevent update from self if instance refers to same record."""
        if self.sourceobject is not None:
            if (
                instance.key == self.sourceobject.key
                and self.datasource.dbname == self.sourceobject.dbname
                and self.datasource.dbset == self.sourceobject.dbset
            ):
                self.blockchange = True

    def get_text_for_statusbar(self):
        """ """
        return "".join(
            (
                "Please wait while finding games for repertoire position in ",
                self.pgn.tags.get(TAG_OPENING, "?"),
            )
        )

    def get_selection_text_for_statusbar(self):
        """ """
        return self.pgn.tags.get(TAG_OPENING, "?")

    def generate_popup_navigation_maps(self):
        navigation_map = {k: v for k, v in self.get_navigation_events()}
        local_map = {
            EventSpec.scoresheet_to_analysis: self.analysis_current_item,
        }
        return navigation_map, local_map


class RepertoireDisplay(
    _RepertoireDisplay, DisplayPGN, ShowPGN, Repertoire, DataNotify
):

    """Display a repertoire from a database allowing delete and insert."""

    # Allow for structure difference between RepertoireDisplay and GameDisplay
    # versions of delete_game_database.
    # Method comments suggest a problem exists which needs fixing.
    # Existence of this method prevents delete_game_database being used by
    # instances of superclasses of RepertoireDisplay, emulating the behaviour
    # before introduction of displaypgn module.
    def pgn_score_original_value(self, original_value):

        # currently attracts "AttributeError: 'ChessDBvalueGameTags' has
        # no attribute 'gamesource'.
        # original.value.set_game_source(self.sourceobject.value.gamesource)
        original_value.set_game_source("No opening name")

    def create_primary_activity_popup(self):
        popup = super().create_primary_activity_popup()
        self.add_close_item_entry_to_popup(popup)
        return popup

    def create_select_move_popup(self):
        popup = super().create_select_move_popup()
        self.add_close_item_entry_to_popup(popup)
        return popup


class RepertoireDisplayInsert(
    _RepertoireDisplay, InsertPGN, ShowPGN, RepertoireEdit, DataNotify
):

    """Display a repertoire from a database allowing insert.

    RepertoireEdit provides the widget and _RepertoireDisplay the database
    interface.

    """


class RepertoireDisplayEdit(EditPGN, RepertoireDisplayInsert):

    """Display a repertoire from a database allowing edit and insert."""

    # Allow for structure difference between RepertoireDisplay and GameDisplay
    # versions of delete_game_database.
    # Method comments suggest a problem exists which needs fixing.
    # Existence of this method prevents delete_game_database being used by
    # instances of superclasses of RepertoireDisplay, emulating the behaviour
    # before introduction of displaypgn module.
    def pgn_score_original_value(self, original_value):

        # currently attracts "AttributeError: 'ChessDBvalueGameTags' has
        # no attribute 'gamesource'.
        # original.value.set_game_source(self.sourceobject.value.gamesource)
        original_value.set_game_source("No opening name")

    # set_properties_on_grids defined so update_game_database method can be
    # shared by repertoiredisplay.RepertoireDisplayEdit and
    # gamedisplay.GameDisplayEdit classes.
    # See class attributes pgn_score_name and pgn_score_source_name too.
    # The property which returns self.ui.base_repertoires is ignored because
    # the GameDisplayEdit version of the method sets properties on all grids.
    def set_properties_on_game_grids(self, newkey):
        self.ui.base_repertoires.set_properties(newkey)



# querystatement.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Game selection rule parser.
"""

import re

from pgn_read.core.constants import TAG_WHITE, TAG_BLACK

from solentware_base.core.where import LIKE

from .constants import NAME_DELIMITER

# Normalize player names for index consistency.
# Format defined in the PGN specification is 'surname, forename I.J.' assuming
# no spaces between initials when more than one are adjacent. Real PGN files do
# not always adhere.
# Names like 'surname, I. forename J.' may occur.
# Index values are created by replacing all sequences containing just commas,
# periods, and whitespace, by a single space.
re_normalize_player_name = re.compile("([^,\.\s]+)(?:[,\.\s]*)")


class QueryStatementError(Exception):
    pass


class QueryStatement(object):
    """Game selection rule parser.

    Parse text for a game selection rule specification.

    """

    def __init__(self):
        """ """
        super().__init__()
        self._dbset = None

        # Support using where.Where or where_dpt.Where depending on database
        # engine being used.
        # This attribute should not be used for anything else.
        self.__database = None

        self._description_string = ""
        self._query_statement_string = ""
        self._where_error = None

        self._reset_rule_state()

    def _reset_rule_state(self):
        """Initialiase the selection rule parsing state.

        The analyser is able to have at least two goes at a statement.

        """
        self.where = None
        self.textok = ""
        self.texterror = ""

    @property
    def where_error(self):
        return self._where_error

    @property
    def dbset(self):
        return self._dbset

    @dbset.setter
    def dbset(self, value):
        if self._dbset is None:
            self._dbset = value
        elif self._dbset != value:
            raise QueryStatementError(
                "".join(
                    (
                        "Database file name already set to ",
                        repr(self._dbset),
                        ", cannot change to ",
                        repr(value),
                        ".",
                    )
                )
            )

    def process_query_statement(self, text):
        """Process selection rule in text.

        First attempt treats whole text as a selection rule.
        Second attempt treats first line, of at least two, as the name of the
        selection rule and the rest as a selection rule.

        """
        if self.__database is None:
            return

        # Assume no error, but set False indicating process_query_statement
        # has been called.
        self._where_error = False

        for rule in (
            ("", text.strip()),
            [t.strip() for t in text.split(NAME_DELIMITER, 1)],
        ):

            if len(rule) == 1:

                # The second element of rule is being processed and the text in
                # rule[1] cannot be a valid selection rule because it is the
                # text which was processed from the first element of rule.
                # self._where_error is bound to a WhereStatementError instance.
                return False

            self._reset_rule_state()
            self._description_string, self._query_statement_string = rule
            w = self.__database.record_selector(self._query_statement_string)
            w.lex()
            w.parse()
            if w.validate(self.__database, self.dbset):
                self._where_error = w.error_information
                continue
            self.where = w
            self.textok = self._query_statement_string
            self.texterror = ""
            self._where_error = False
            for n in w.node.get_clauses_from_root_in_walk_order():
                if n.field in (TAG_WHITE, TAG_BLACK):
                    if n.condition == LIKE:
                        continue
                    if not isinstance(n.value, tuple):
                        n.value = " ".join(
                            re_normalize_player_name.findall(n.value)
                        )
                    else:
                        n.value = tuple(
                            [
                                " ".join(re_normalize_player_name.findall(nv))
                                for nv in n.value
                            ]
                        )
            return True

        # self._where_error is not bound to a WhereStatementError instance.
        return False

    def get_name_text(self):
        """Return name text."""
        return self._description_string

    def get_name_query_statement_text(self):
        """Return name and position text."""
        return NAME_DELIMITER.join(
            (
                self._description_string,
                self._query_statement_string,
            )
        )

    def get_query_statement_text(self):
        """Return position text."""
        return self._query_statement_string

    def set_database(self, database=None):
        """Set Database instance to which selection rule is applied."""
        self.__database = database
